/* Emacs style mode select   -*- C++ -*-
 *-----------------------------------------------------------------------------
 *
 *
 *  PrBoom: a Doom port merged with LxDoom and LSDLDoom
 *  based on BOOM, a modified and improved DOOM engine
 *  Copyright (C) 1999 by
 *  id Software, Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman
 *  Copyright (C) 1999-2006 by
 *  Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze
 *  Copyright 2005, 2006 by
 *  Florian Schulze, Colin Phipps, Neil Stevens, Andrey Budko
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * DESCRIPTION:
 *  DOOM graphics stuff for NDS
 *
 *-----------------------------------------------------------------------------
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdint.h>

typedef  uint8_t Uint8;

#include <nds.h>

#include "m_argv.h"
#include "doomstat.h"
#include "doomdef.h"
#include "doomtype.h"
#include "v_video.h"
#include "r_draw.h"
#include "d_main.h"
#include "d_event.h"
#include "i_joy.h"
#include "i_video.h"
#include "z_zone.h"
#include "s_sound.h"
#include "sounds.h"
#include "w_wad.h"
#include "st_stuff.h"
#include "lprintf.h"
#include "i_pngshot.h"

int gl_colorbuffer_bits=16;
int gl_depthbuffer_bits=16;

extern void M_QuitDOOM(int choice);
#ifdef DISABLE_DOUBLEBUFFER
int use_doublebuffer = 0;
#else
int use_doublebuffer = 1; // Included not to break m_misc, but not relevant to SDL
#endif
int use_fullscreen;
int desired_fullscreen;
static SDL_Surface *screen;

////////////////////////////////////////////////////////////////////////////
// Input code
int             leds_always_off = 0; // Expected by m_misc, not relevant

// Mouse handling
extern int     usemouse;        // config file var
static boolean mouse_enabled; // usemouse, but can be overriden by -nomouse
static boolean mouse_currently_grabbed;




struct eventTranslate {
  u32 buttonMask;
  u32 doomKey;
};

struct eventTranslate gameKeyTable[] = {
  { KEY_UP     , KEYD_UPARROW    },
  { KEY_DOWN   , KEYD_DOWNARROW  },
  { KEY_LEFT   , KEYD_LEFTARROW  },
  { KEY_RIGHT  , KEYD_RIGHTARROW },
  { KEY_START  , KEYD_ESCAPE     },
  { KEY_SELECT , KEYD_TAB        },
  { KEY_A      , KEYD_RCTRL      },
  { KEY_B      , KEYD_SPACEBAR   },
  { KEY_L      , ','             },
  { KEY_R      , '.'             },
  { KEY_X      , KEYD_RSHIFT     }
};

int numGameKeys = sizeof(gameKeyTable) / sizeof(gameKeyTable[0]);

struct eventTranslate menuKeyTable[] = {
  { KEY_UP     , KEYD_UPARROW    },
  { KEY_DOWN   , KEYD_DOWNARROW  },
  { KEY_LEFT   , KEYD_LEFTARROW  },
  { KEY_RIGHT  , KEYD_RIGHTARROW },
  { KEY_START  , KEYD_ESCAPE     },
  { KEY_SELECT , KEYD_ENTER      },
  { KEY_A      , KEYD_ENTER      },
  { KEY_B      , KEYD_BACKSPACE  },
  { KEY_Y      , 'y'             }
};

int numMenuKeys = sizeof(menuKeyTable) / sizeof(menuKeyTable[0]);

struct eventTranslate mapKeyTable[] = {
  { KEY_UP     , KEYD_UPARROW    },
  { KEY_DOWN   , KEYD_DOWNARROW  },
  { KEY_LEFT   , KEYD_LEFTARROW  },
  { KEY_RIGHT  , KEYD_RIGHTARROW },
  { KEY_SELECT , KEYD_TAB        },
  { KEY_START  , KEYD_ESCAPE     },
  { KEY_L      , '-'             },
  { KEY_R      , '='             }
};

int numMapKeys = sizeof(mapKeyTable) / sizeof(mapKeyTable[0]);

void translateKeys(evtype_t type, u32 mask, struct eventTranslate *table, int count)
{
  int i;

  event_t event;
  event.type = type;

  for( i=0; i<count; i++)
  {
    if(mask & table[i].buttonMask)
    {
      event.data1 = table[i].doomKey;
      D_PostEvent(&event);
    }
  }
}

char weapons[9] = { '1', '2', '3', '4', '5', '6', '7', '8', '9' };
int last_weapon_key = 0;


void gameControls() {

  u32 kDown, kUp;

  struct eventTranslate *translateTable = gameKeyTable;
  int numTranslations = numGameKeys;

  kDown = keysDown();
  kUp   = keysUp();

  if (menuactive)
  {
    translateTable = menuKeyTable;
    numTranslations = numMenuKeys;
  } else if (automapmode & am_active) {
    translateTable = mapKeyTable;
    numTranslations = numMapKeys;
  } else {

    event_t event;

    if (last_weapon_key != 0 && players[displayplayer].pendingweapon != wp_nochange) {
        event.type = ev_keyup;
        event.data1 = last_weapon_key;
        D_PostEvent(&event);
        last_weapon_key = 0;
    }


    if(kDown & KEY_Y) {

        weapontype_t weapon_index = players[displayplayer].readyweapon;

        weapon_index++;
        while(players[displayplayer].weaponowned[weapon_index] == false) {
            weapon_index++;
            if (weapon_index >= NUMWEAPONS) weapon_index = 0;
        }

        last_weapon_key = weapons[weapon_index];
        event.type = ev_keydown;
        event.data1 = last_weapon_key;
        D_PostEvent(&event);
    }
  }

  translateKeys(ev_keydown, kDown, translateTable, numTranslations);

  translateKeys(ev_keyup, kUp, translateTable, numTranslations);


}
//
// I_StartTic
//

void I_StartTic (void)
{
  scanKeys();
  gameControls();
}

//
// I_StartFrame
//
void I_StartFrame (void)
{
}

//
// I_InitInputs
//

static void I_InitInputs(void)
{
  int nomouse_parm = M_CheckParm("-nomouse");

  // check if the user wants to use the mouse
  mouse_enabled = usemouse && !nomouse_parm;

}

///////////////////////////////////////////////////////////
// Palette stuff.
//
typedef struct
{
  u8 r;
  u8 g;
  u8 b;
  u8 unused;
} SDL_Color;


static void I_UploadNewPalette(int pal)
{
  // This is used to replace the current 256 colour cmap with a new one
  // Used by 256 colour PseudoColor modes

  // Array of SDL_Color structs used for setting the 256-colour palette
  static SDL_Color* colours;
  static int cachedgamma;
  static size_t num_pals;

  if ((colours == NULL) || (cachedgamma != usegamma)) {
    int pplump = W_GetNumForName("PLAYPAL");
    int gtlump = (W_CheckNumForName)("GAMMATBL",ns_prboom);
    register const byte * palette = W_CacheLumpNum(pplump);
    register const byte * const gtable = (const byte *)W_CacheLumpNum(gtlump) + 256*(cachedgamma = usegamma);
    register int i;

    num_pals = W_LumpLength(pplump) / (3*256);
    num_pals *= 256;

    if (!colours) {
      // First call - allocate and prepare colour array
      colours = malloc(sizeof(SDL_Color)*num_pals);
    }

    // set the colormap entries
    for (i=0 ; (size_t)i<num_pals ; i++) {
      colours[i].r = gtable[palette[0]];
      colours[i].g = gtable[palette[1]];
      colours[i].b = gtable[palette[2]];
      palette += 3;
    }

    W_UnlockLumpNum(pplump);
    W_UnlockLumpNum(gtlump);
    num_pals/=256;
  }
#ifdef RANGECHECK
  if ((size_t)pal >= num_pals) I_Error("I_UploadNewPalette: Palette number out of range (%d>=%d)", pal, num_pals);
#endif

  u32 i;
  for(i = 0; i < 256; i++)
  {
    u8 r, g, b;
    r = (u8)colours[i+256*pal].r;
    g = (u8)colours[i+256*pal].g;
    b = (u8)colours[i+256*pal].b;
    BG_PALETTE[i]=RGB8(r,g,b);
  }

}
//////////////////////////////////////////////////////////////////////////////
// Graphics API

void I_ShutdownGraphics(void)
{
}

//
// I_UpdateNoBlit
//
void I_UpdateNoBlit (void)
{
}

//
// I_FinishUpdate
//
static int newpal = 0;
#define NO_PALETTE_CHANGE 1000

void I_FinishUpdate (void)
{

	int h = 200;
	int w = 320;
	int step = 512;
	unsigned char *srcmain = screens[0].data;
	unsigned char *destmain = NULL;


	destmain = (unsigned char *)BG_GFX;

  for (; h>0; h--)
  {
		dmaCopy(srcmain, destmain, w);
		destmain += step;
		srcmain += SCREENWIDTH;
	}

  // Update the display buffer (flipping video pages if supported)
  // If we need to change palette, that implicitely does a flip
  if (newpal != NO_PALETTE_CHANGE) {
    I_UploadNewPalette(newpal);
    newpal = NO_PALETTE_CHANGE;
  }
}

//
// I_ScreenShot
//

#ifdef HAVE_LIBPNG
#define SAVE_PNG_OR_BMP I_SavePNG
#else
#define SAVE_PNG_OR_BMP SDL_SaveBMP
#endif

int I_ScreenShot (const char *fname)
{
//  return SAVE_PNG_OR_BMP(SDL_GetVideoSurface(), fname);
	return 0;
}

//
// I_SetPalette
//
void I_SetPalette (int pal)
{
  newpal = pal;
}

// I_PreInitGraphics

static void I_Shutdown(void)
{
//  SDL_Quit();
  return;
}

void I_PreInitGraphics(void)
{
}

// CPhipps -
// I_CalculateRes
// Calculates the screen resolution, possibly using the supplied guide
void I_CalculateRes(unsigned int width, unsigned int height)
{
    SCREENWIDTH = width;
    SCREENHEIGHT = height;
    SCREENPITCH = (width * V_GetPixelDepth() + 3) & ~3;
}

// CPhipps -
// I_SetRes
// Sets the screen resolution
void I_SetRes(void)
{
  int i;

  I_CalculateRes(SCREENWIDTH, SCREENHEIGHT);

  // set first three to standard values
  for (i=0; i<1; i++) {
    screens[i].width = SCREENWIDTH;
    screens[i].height = SCREENHEIGHT;
    screens[i].byte_pitch = SCREENPITCH;
    screens[i].short_pitch = SCREENPITCH / V_GetModePixelDepth(VID_MODE16);
    screens[i].int_pitch = SCREENPITCH / V_GetModePixelDepth(VID_MODE32);
  }

  // statusbar
  screens[4].width = SCREENWIDTH;
  screens[4].height = (ST_SCALED_HEIGHT+1);
  screens[4].byte_pitch = SCREENPITCH;
  screens[4].short_pitch = SCREENPITCH / V_GetModePixelDepth(VID_MODE16);
  screens[4].int_pitch = SCREENPITCH / V_GetModePixelDepth(VID_MODE32);

  lprintf(LO_INFO,"I_SetRes: Using resolution %dx%d\n", SCREENWIDTH, SCREENHEIGHT);
}

void I_InitGraphics(void)
{
  char titlebuffer[2048];
  static int    firsttime=1;

  if (firsttime)
  {
    firsttime = 0;

	SCREENWIDTH = 256;
	SCREENHEIGHT = 192;
	SCREENPITCH = 256;

    atexit(I_ShutdownGraphics);
    lprintf(LO_INFO, "I_InitGraphics: %dx%d\n", SCREENWIDTH, SCREENHEIGHT);

    /* Set the video mode */
    I_UpdateVideoMode();

    /* Setup the window title */
    snprintf(titlebuffer,sizeof(titlebuffer),"%s %s", PACKAGE, VERSION);

    /* Initialize the input system */
    I_InitInputs();
  }
}

int I_GetModeFromString(const char *modestr)
{
  return VID_MODE8;
}

void I_UpdateVideoMode(void)
{
  int init_flags;
  int i;
  video_mode_t mode;

  lprintf(LO_INFO, "I_UpdateVideoMode: %dx%d (%s)\n", SCREENWIDTH, SCREENHEIGHT, desired_fullscreen ? "fullscreen" : "nofullscreen");

  mode = I_GetModeFromString(default_videomode);
  if ((i=M_CheckParm("-vidmode")) && i<myargc-1) {
    mode = I_GetModeFromString(myargv[i+1]);
  }

  V_InitMode(mode);
  V_DestroyUnusedTrueColorPalettes();
  V_FreeScreens();

  I_SetRes();

  screens[0].not_on_heap = false;
  screens[0].width = SCREENWIDTH;
  screens[0].height = SCREENHEIGHT;
  screens[0].byte_pitch = SCREENPITCH;
  screens[0].short_pitch = screens[0].byte_pitch/2;
  screens[0].int_pitch = screens[0].short_pitch/2;

  V_AllocScreens();
  R_InitBuffer(SCREENWIDTH, SCREENHEIGHT);
}
